# Contributor information:
#   Name: First Last
#   GitHub account: username
#   Email: username@abcd.edu
#   Organization: ABCD University
"""This is a template for community contributions."""


def example_function():
  """This is an example function."""
  return True
